from ._Transform import TransformInteractivity
from ._Rotate import RotateTransformInteractivity
from ._Scale import ScaleTransformInteractivity
from ._Skew import SkewTransformInteractivity
from ._Translate import TranslateTransformInteractivity

__all__ = (
    "TransformInteractivity",
    "TranslateTransformInteractivity",
    "SkewTransformInteractivity",
    "RotateTransformInteractivity",
    "ScaleTransformInteractivity"
)



from enum import Enum, unique
from typing import Any, Dict, Optional, Tuple, Union, cast

from PyQt5.QtCore import QPointF, QSizeF, pyqtSignal
from PyQt5.QtWidgets import QInputDialog, QWidget

import ImagingS.Gui.ui as ui
from ImagingS.brush import Brush
from ImagingS.document import Document
from ImagingS.drawing import Drawing, DrawingGroup, GeometryDrawing, Pen
from ImagingS.geometry import (CurveAlgorithm, CurveGeometry, EllipseGeometry,
                               Geometry, LineAlgorithm, LineGeometry,
                               LineClipAlgorithm, LineClipper,
                               PolygonGeometry, PolylineGeometry,
                               RectangleGeometry)
from ImagingS.Gui import converters, icons
from ImagingS.Gui.graphic import Canvas
from ImagingS.Gui.interactivity import Interactivity, InteractivityState, RectClipInteractivity
from ImagingS.Gui.interactivity.geometry import (CurveInteractivity,
                                                 EllipseInteractivity,
                                                 GeometryInteractivity,
                                                 LineInteractivity,
                                                 PolylineInteractivity,
                                                 RectangleInteractivity)
from ImagingS.Gui.interactivity.transform import (
    RotateTransformInteractivity, ScaleTransformInteractivity,
    SkewTransformInteractivity, TransformInteractivity,
    TranslateTransformInteractivity)
from ImagingS.transform import (MatrixTransform, RotateTransform,
                                ScaleTransform, SkewTransform, TransformGroup,
                                TranslateTransform)

_TransformInteractiveCreateMap: Dict[str, Tuple[Any, Any]] = {
    "Translate": (TranslateTransform, TranslateTransformInteractivity),
    "Rotate": (RotateTransform, RotateTransformInteractivity),
    "Scale": (ScaleTransform, ScaleTransformInteractivity),
    "Skew": (SkewTransform, SkewTransformInteractivity),
}


@unique
class VisualPageState(Enum):
    Disable = 0
    Normal = 1
    Interactive = 2


class VisualPage(QWidget, ui.VisualPage):
    messaged = pyqtSignal(str)
    documentCommitted = pyqtSignal(Document, str)
    stateChanged = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self.setupToolBar()
        self.setupIcon()
        self.setupCanvas()

        # self.actionDrawings = [
        #     self.actDrawingLine,
        #     self.actDrawingPolygon,
        #     self.actDrawingCurve,
        #     self.actDrawingEllipse,
        # ]

        for act in [self.actTransformTranslate,
                    self.actTransformScale,
                    self.actTransformRotate,
                    self.actTransformSkew]:
            act.triggered.connect(self.actTransformCreate_triggered)

        self.actTransformMatrix.triggered.connect(self.actTransformMatrix_triggered)
        self.actTransformGroup.triggered.connect(self.actTransformGroup_triggered)

        self.actDrawingLine.triggered.connect(
            self.actDrawingLine_triggered)
        self.actDrawingPolygon.triggered.connect(
            self.actDrawingPolygon_triggered)
        self.actDrawingPolyline.triggered.connect(
            self.actDrawingPolyline_triggered)
        self.actDrawingRectangle.triggered.connect(
            self.actDrawingRectangle_triggered)
        self.actDrawingCurve.triggered.connect(
            self.actDrawingCurve_triggered)
        self.actDrawingEllipse.triggered.connect(
            self.actDrawingEllipse_triggered)

        self.actClip.triggered.connect(self.actClip_triggered)

        self.setEnabled(False)
        self._drawing = None
        self._state = VisualPageState.Disable

    def setupToolBar(self):
        # tb = QToolButton()
        # tb.setDefaultAction(self.actDrawingLine)
        # tb.setPopupMode(QToolButton.DelayedPopup)
        # self.tlbMain.addWidget(tb)

        # tb = QToolButton()
        # menu = QMenu()
        # menu.addActions([self.actDrawingPolygon,
        #                  self.actDrawingPolyline,
        #                  self.actDrawingRectangle])
        # tb.setDefaultAction(self.actDrawingPolygon)
        # tb.setMenu(menu)
        # tb.setPopupMode(QToolButton.DelayedPopup)
        # self.tlbMain.addWidget(tb)

        # tb = QToolButton()
        # tb.setDefaultAction(self.actDrawingCurve)
        # tb.setPopupMode(QToolButton.DelayedPopup)
        # self.tlbMain.addWidget(tb)

        # tb = QToolButton()
        # tb.setDefaultAction(self.actDrawingEllipse)
        # self.tlbMain.addWidget(tb)

        # self.tlbMain.addSeparator()

        # self.tlbMain.addActions([
        #     self.actTransformTranslate,
        #     self.actTransformScale,
        #     self.actTransformRotate,
        #     self.actTransformSkew,
        #     self.actTransformMatrix,
        #     self.actTransformGroup,
        # ])
        pass

    def setupIcon(self):
        self.actDrawingLine.setIcon(icons.lineGeometry)
        self.actDrawingCurve.setIcon(icons.curveGeometry)
        self.actDrawingEllipse.setIcon(icons.ellipseGeometry)
        self.actDrawingPolygon.setIcon(icons.polygonGeometry)
        self.actDrawingPolyline.setIcon(icons.polylineGeometry)
        self.actDrawingRectangle.setIcon(icons.rectangleGeometry)
        self.actTransformSkew.setIcon(icons.skewTransform)
        self.actTransformScale.setIcon(icons.scaleTransform)
        self.actTransformTranslate.setIcon(icons.translateTransform)
        self.actTransformRotate.setIcon(icons.rotateTransform)
        self.actTransformMatrix.setIcon(icons.matrixTransform)
        self.actTransformGroup.setIcon(icons.transformGroup)
        self.actClip.setIcon(icons.clip)

    def setupCanvas(self):
        self.cvsMain = Canvas(self.widMain)
        self.cvsMain.setObjectName("cvsMain")
        self.grdMain.addWidget(self.cvsMain, 0, 0, 1, 1)
        self.cvsMain.resize(QSizeF(600, 600))
        self.cvsMain.mouseMoved.connect(
            self.cvsMain_mouseMoved)
        self.cvsMain.mouseReleased.connect(
            self.cvsMain_mouseReleased)

    def enable(self, doc: Document) -> None:
        assert self.state is VisualPageState.Disable
        self.setEnabled(True)
        self._document = doc
        self.drawing = None
        self.brush = None
        self._setState(VisualPageState.Normal)
        self.fresh()

    def disable(self) -> None:
        assert self.state is not VisualPageState.Disable
        if hasattr(self, "_document"):
            del self._document
        self.drawing = None
        self.brush = None
        self.cvsMain.clear()
        self.setEnabled(False)
        self._setState(VisualPageState.Disable)

    def _setState(self, value: VisualPageState) -> None:
        self._state = value

        if self._state is VisualPageState.Interactive:
            self.tlbMain.setEnabled(False)
        elif self._state is VisualPageState.Normal:
            self.tlbMain.setEnabled(True)

        self.stateChanged.emit()

    @property
    def state(self) -> VisualPageState:
        return self._state

    @property
    def drawing(self) -> Optional[Drawing]:
        return self._drawing

    @drawing.setter
    def drawing(self, value: Optional[Drawing]) -> None:
        self._drawing = value
        if self.state is VisualPageState.Normal:
            self.cvsMain.select(
                self._drawing.id if self._drawing is not None else None)

    @property
    def brush(self) -> Optional[Brush]:
        return self._brush

    @brush.setter
    def brush(self, value: Optional[Brush]) -> None:
        self._brush = value

    def fresh(self) -> None:
        if self.state is not VisualPageState.Normal:
            return

        self.cvsMain.clear()
        self.cvsMain.resize(converters.qsize(self._document.size))
        for dr in self._document.drawings.children:
            self.cvsMain.add(dr)

    def cvsMain_mouseMoved(self, point: QPointF):
        self.messaged.emit(str((round(point.x()), round(point.y()))))

    def cvsMain_mouseReleased(self, point: QPointF):
        if self.state is VisualPageState.Normal:
            p = converters.point(point)
            for dr in reversed(self._document.drawings.children.items):
                if p in dr.bounds:
                    self.drawing = dr
                    break

    def _emptyGeometryDrawing(self) -> GeometryDrawing:
        drawing = GeometryDrawing()
        if self.brush is not None:
            drawing.stroke = Pen(self.brush)
        return drawing

    def _beginInteractive(self, inter: Interactivity) -> None:
        inter.ended.connect(self.interactivity_ended)
        inter.started.connect(self.interactivity_started)
        inter.updated.connect(self.interactivity_updated)
        self.cvsMain.interactivity = inter
        inter.start()

    def interactivity_started(self, inter: Interactivity) -> None:
        self._setState(VisualPageState.Interactive)
        if inter.viewItem is not None:
            self.cvsMain.scene().addItem(inter.viewItem)
        self.cvsMain.rerender()

    def interactivity_updated(self, inter: Interactivity) -> None:
        self.cvsMain.rerender()

    def interactivity_ended(self, inter: Interactivity) -> None:
        self._setState(VisualPageState.Normal)
        if inter.viewItem is not None:
            self.cvsMain.scene().removeItem(inter.viewItem)
        self.cvsMain.interactivity = None
        if inter.state is InteractivityState.Success:
            if isinstance(inter, GeometryInteractivity):
                drawing = cast(GeometryInteractivity, inter).target
                self._document.drawings.children.append(drawing)
                self.documentCommitted.emit(
                    self._document, f"Create {inter.geometry.__class__.__name__}")
            elif isinstance(inter, TransformInteractivity):
                if self.drawing is not None:
                    self.documentCommitted.emit(
                        self._document, f"Create {inter.transform.__class__.__name__}")
            elif isinstance(inter, RectClipInteractivity):
                if self.drawing is not None:
                    if isinstance(self.drawing, GeometryDrawing) and isinstance(self.drawing.geometry, LineGeometry):
                        drawingId = self.drawing.id

                        geo = self.drawing.geometry
                        clipper = LineClipper(cast(LineGeometry, geo.transformed()))
                        clipped = None
                        if inter.clip is not None:
                            clipped = clipper.clip(inter.clip, inter.clipAlgorithm)

                        if clipped is None:
                            parent = self.drawing.parent()
                            if parent is None:
                                return
                            del parent[drawingId]
                        else:
                            geo.start = clipped.start
                            geo.end = clipped.end
                            geo.transform = None  # clear transform

                        self.documentCommitted.emit(
                            self._document, f"Clip {drawingId}")
        self.fresh()

    def actDrawingLine_triggered(self):
        algs = [e.name for e in LineAlgorithm]
        item, okPressed = QInputDialog.getItem(
            self, "Select Algorithm", "Algorithm:", algs, 0, False)
        if not okPressed or not item:
            return
        geo = LineGeometry()
        geo.algorithm = getattr(LineAlgorithm, item)
        self._beginInteractive(LineInteractivity(
            self._emptyGeometryDrawing(), geo, self._document.size))

    def actDrawingPolygon_triggered(self):
        algs = [e.name for e in LineAlgorithm]
        item, okPressed = QInputDialog.getItem(
            self, "Select Algorithm", "Algorithm:", algs, 0, False)
        if not okPressed or not item:
            return
        geo = PolygonGeometry()
        geo.algorithm = getattr(LineAlgorithm, item)
        self._beginInteractive(PolylineInteractivity(
            self._emptyGeometryDrawing(), geo, self._document.size))

    def actDrawingPolyline_triggered(self):
        algs = [e.name for e in LineAlgorithm]
        item, okPressed = QInputDialog.getItem(
            self, "Select Algorithm", "Algorithm:", algs, 0, False)
        if not okPressed or not item:
            return
        geo = PolylineGeometry()
        geo.algorithm = getattr(LineAlgorithm, item)
        self._beginInteractive(PolylineInteractivity(
            self._emptyGeometryDrawing(), geo, self._document.size))

    def actDrawingRectangle_triggered(self):
        algs = [e.name for e in LineAlgorithm]
        item, okPressed = QInputDialog.getItem(
            self, "Select Algorithm", "Algorithm:", algs, 0, False)
        if not okPressed or not item:
            return
        geo = RectangleGeometry()
        geo.algorithm = getattr(LineAlgorithm, item)
        self._beginInteractive(RectangleInteractivity(
            self._emptyGeometryDrawing(), geo, self._document.size))

    def actDrawingCurve_triggered(self):
        algs = [e.name for e in CurveAlgorithm]
        item, okPressed = QInputDialog.getItem(
            self, "Select Algorithm", "Algorithm:", algs, 0, False)
        if not okPressed or not item:
            return
        geo = CurveGeometry()
        geo.algorithm = getattr(CurveAlgorithm, item)
        self._beginInteractive(CurveInteractivity(
            self._emptyGeometryDrawing(), geo, self._document.size))

    def actDrawingEllipse_triggered(self):
        geo = EllipseGeometry()
        self._beginInteractive(EllipseInteractivity(
            self._emptyGeometryDrawing(), geo, self._document.size))

    def actClip_triggered(self):
        if self.drawing is None:
            return
        target = None
        if isinstance(self.drawing, DrawingGroup):
            # target = self.drawing TODO
            pass
        elif isinstance(self.drawing, GeometryDrawing) and isinstance(self.drawing.geometry, LineGeometry):
            target = self.drawing.geometry
        if target is None:
            return
        algs = [e.name for e in LineClipAlgorithm]
        item, okPressed = QInputDialog.getItem(
            self, "Select Algorithm", "Algorithm:", algs, 0, False)
        if not okPressed or not item:
            return
        rectClipInter = RectClipInteractivity(
            target, self._document.size)
        rectClipInter.clip = None
        rectClipInter.clipAlgorithm = getattr(LineClipAlgorithm, item)
        self._beginInteractive(rectClipInter)

    def _getTransformInteractiveTarget(self) -> Optional[Union[DrawingGroup, Geometry]]:
        if self.drawing is None:
            return
        target = None
        if isinstance(self.drawing, DrawingGroup):
            target = self.drawing
        elif isinstance(self.drawing, GeometryDrawing):
            target = self.drawing.geometry
        if target is None:
            return None
        return target

    def actTransformCreate_triggered(self):
        target = self._getTransformInteractiveTarget()
        if target is None:
            return

        sender = self.sender()
        tname = sender.objectName()[len("actTransform"):]
        assert tname in _TransformInteractiveCreateMap
        transType, interType = _TransformInteractiveCreateMap[tname]

        if not isinstance(target.transform, (type(None), TransformGroup)):
            self.actTransformGroup.trigger()

        tr = transType.__call__()
        self._beginInteractive(interType.__call__(target, tr))

    def actTransformGroup_triggered(self):
        target = self._getTransformInteractiveTarget()
        if target is None:
            return
        if target.transform is None:
            target.transform = TransformGroup()
        elif not isinstance(target.transform, TransformGroup):
            gr = TransformGroup()
            gr.children.append(target.transform)
            target.transform = gr
        else:
            return
        self.documentCommitted.emit(self._document, f"Create TransformGroup")
        self.fresh()

    def actTransformMatrix_triggered(self):
        target = self._getTransformInteractiveTarget()
        if target is None:
            return
        if not isinstance(target.transform, (type(None), TransformGroup)):
            self.actTransformGroup.trigger()
        trans = MatrixTransform()
        default = "\n".join([", ".join(map(str, c))
                             for c in trans.matrix.tolist()])
        text, okPressed = QInputDialog.getMultiLineText(
            self, "Input Matrix", "Matrix:", default)
        if not okPressed or not text:
            return
        lines = str(text).strip().splitlines()
        try:
            for i in range(3):
                t = lines[i].split(",")
                for j in range(3):
                    trans.matrix[i][j] = float(t[j])
        except Exception:
            self.messaged.emit("Matrix is invalid.")
            return
        if isinstance(target.transform, TransformGroup):
            target.transform.children.append(trans)
        else:
            target.transform = trans
        self.documentCommitted.emit(self._document, f"Create MatrixTransform")
        self.fresh()



from typing import Optional, Union

from ImagingS.geometry import LineGeometry, LineClipAlgorithm
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsPathItem
from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QColor, QKeyEvent, QPainterPath, QPen

from ImagingS import Rect, Size
from ImagingS.drawing import DrawingGroup
from ImagingS.Gui import converters

from . import Interactivity


class RectClipInteractivity(Interactivity):
    def __init__(self, target: Union[DrawingGroup, LineGeometry], size: Size) -> None:
        super().__init__()
        self._target = target

        self._viewItem = QGraphicsPathItem()
        pen = QPen(QColor("purple"))
        pen.setWidth(1)
        pen.setStyle(Qt.DashLine)
        pen.setCapStyle(Qt.RoundCap)
        self._viewItem.setPen(pen)

    @property
    def viewItem(self) -> Optional[QGraphicsItem]:
        return self._viewItem

    @property
    def target(self) -> Union[DrawingGroup, LineGeometry]:
        return self._target

    @property
    def clip(self) -> Optional[Rect]:
        return self._clip

    @clip.setter
    def clip(self, value: Optional[Rect]) -> None:
        self._clip = value

    @property
    def clipAlgorithm(self) -> LineClipAlgorithm:
        return self._clipAlgorithm

    @clipAlgorithm.setter
    def clipAlgorithm(self, value: LineClipAlgorithm) -> None:
        self._clipAlgorithm = value

    def start(self) -> None:
        self._hasStarted = False
        self._isShift = False
        super().start()

    def update(self) -> None:
        self.target.refreshBounds()
        super().update()

    def _updateData(self) -> None:
        if isinstance(self.target, LineGeometry):
            # self.target.clip = Rect.fromPoints(self._start, self._end)
            pass
        else:
            assert False  # not support generic clip for now TODO
        path = QPainterPath()
        path.addRect(converters.qrect(Rect.fromPoints(self._start, self._end)))
        self._viewItem.setPath(path)

    def end(self, success: bool) -> None:
        if not success:
            if isinstance(self.target, LineGeometry):
                self.clip = None
            else:
                self.target.clip = None
        else:
            if isinstance(self.target, LineGeometry):
                self.clip = Rect.fromPoints(self._start, self._end)
            else:
                self.target.clip = Rect.fromPoints(self._start, self._end)
            
        super().end(success)

    def _setEndPoint(self, point: QPointF) -> None:
        if self._isShift:
            tp = converters.point(point)
            delta = tp - self._start
            if abs(delta.x) <= abs(delta.y):
                delta.y = delta.x
            else:
                delta.x = delta.y
            self._end = self._start + delta
        else:
            self._end = converters.point(point)

    def onMouseRelease(self, point: QPointF) -> None:
        if not self._hasStarted:
            self._start = converters.point(point)
            self._hasStarted = True
        else:
            self._setEndPoint(point)
            self._updateData()
            self.end(True)
        super().onMouseRelease(point)

    def onMouseMove(self, point: QPointF) -> None:
        if self._hasStarted:
            self._setEndPoint(point)
            self._updateData()
            self.update()
        super().onMouseMove(point)

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = True
        if key.key() == Qt.Key_Escape:
            self.end(False)
        super().onKeyPress(key)

    def onKeyRelease(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = False
        super().onKeyRelease(key)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, List, Optional

from ImagingS import Point, Rect, RectMeasurer
from ImagingS.drawing import Pen
from ImagingS.serialization import PropertySerializable
from ImagingS.transform import Transform


class Geometry(PropertySerializable, ABC):
    def __init__(self) -> None:
        super().__init__()
        self.transform = None

    @property
    def transform(self) -> Optional[Transform]:
        return self._transform

    @transform.setter
    def transform(self, value: Optional[Transform]) -> None:
        assert isinstance(value, (type(None), Transform))
        self._transform = value
        self.refreshBounds()

    @abstractmethod
    def strokePoints(self, pen: Pen) -> Iterable[Point]: pass

    @abstractmethod
    def fillPoints(self) -> Iterable[Point]: pass

    def inStroke(self, pen: Pen, point: Point) -> bool:
        return point in self.strokePoints(pen)

    def inFill(self, point: Point) -> bool:
        return point in self.fillPoints()

    @abstractmethod
    def transformed(self) -> Geometry: pass

    @property
    def bounds(self) -> Rect:
        if not hasattr(self, "_bounds"):
            if hasattr(self, "_measurering"):
                return Rect()
            self._measurering = True
            self._bounds = self._calculateBounds()
            del self._measurering
        return self._bounds

    def _calculateBounds(self) -> Rect:
        measurer = RectMeasurer()
        for p in self.strokePoints(Pen()):
            measurer.append(p)
        return measurer.result()

    def refreshBounds(self) -> None:
        if hasattr(self, "_bounds"):
            del self._bounds


class GeometryGroup(Geometry):
    def __init__(self):
        super().__init__()
        self.children = []

    @property
    def children(self) -> List[Geometry]:
        return self._children

    @children.setter
    def children(self, value: List[Geometry]) -> None:
        assert isinstance(value, list)
        self._children = value

    def transformed(self) -> Geometry:
        if self.transform is None:
            return self
        else:
            result = GeometryGroup()
            for c in self.children:
                result.children.append(c.transformed())
            return result

    def strokePoints(self, pen: Pen) -> Iterable[Point]:
        for child in self.children:
            for point in child.strokePoints(pen):
                if self.transform is None:
                    yield point
                else:
                    yield self.transform.transform(point)

    def fillPoints(self) -> Iterable[Point]:
        for child in self.children:
            for point in child.fillPoints():
                if self.transform is None:
                    yield point
                else:
                    yield self.transform.transform(point)

    def inStroke(self, pen: Pen, point: Point) -> bool:
        for child in self.children:
            if child.inStroke(pen, point):
                return True
        return False

    def inFill(self, point: Point) -> bool:
        for child in self.children:
            if child.inFill(point):
                return True
        return False



import math

import numpy as np

from ImagingS import Point, feq
from ImagingS.transform import (MatrixTransform, RotateTransform,
                                ScaleTransform, SkewTransform,
                                TranslateTransform)


def test_clip() -> None:
    pass


def test_matrix() -> None:
    tr = MatrixTransform(np.array([[1, 2, 0], [3, 4, 0], [0, 0, 1]]))
    assert tr.matrix[0][0] == 1
    assert Point(1, 3) == tr.transform(Point(1, 0))
    assert Point(2, 4) == tr.transform(Point(0, 1))


def test_rotate() -> None:
    tr = RotateTransform(Point(0, 0), math.pi / 2)
    assert Point(0, 0) == tr.center
    assert feq(tr.angle, math.pi / 2)
    assert Point(0, 1) == tr.transform(Point(1, 0))
    assert Point(-1, 0) == tr.transform(Point(0, 1))


def test_scale() -> None:
    tr = ScaleTransform(Point(0, 0), (2, 2))
    assert Point(0, 0) == tr.center
    assert (2, 2) == tr.factor
    assert Point(2, 0) == tr.transform(Point(1, 0))
    assert Point(0, 2) == tr.transform(Point(0, 1))


def test_skew() -> None:
    tr = SkewTransform(Point(0, 0), (math.pi / 4, 0))
    assert Point(0, 0) == tr.center
    assert feq(tr.angle[0], math.pi / 4)
    assert feq(tr.angle[1], 0)
    assert Point(1, 0) == tr.transform(Point(1, 0))
    assert Point(1, 1) == tr.transform(Point(0, 1))


def test_translate() -> None:
    tr = TranslateTransform(Point(1, 1))
    assert Point(1, 1) == tr.delta
    assert Point(2, 1) == tr.transform(Point(1, 0))
    assert Point(1, 2) == tr.transform(Point(0, 1))



from __future__ import annotations

from typing import Iterable, Iterator, Optional, Tuple, cast

from ImagingS import Point, Rect
from ImagingS.drawing import Pen

from . import Geometry


def _gen(center: Point, radius: Tuple[float, float]) -> Iterator[Point]:
    def genFirst(a: int, b: int) -> Iterator[Point]:
        if a <= 2 or b <= 2:
            return
        x, y = 0, b
        a2, b2 = a**2, b**2
        d = b2 - a2*(b-0.25)
        yield Point(x, y)
        sp = a2 / (a2+b2)**0.5
        while x < sp:
            if d < 0:
                d += b2*(2*x+3)
            else:
                d += b2*(2*x+3)+a2*(-2*y+2)
                y -= 1
            x += 1
            yield Point(x, y)
        d = b2*(x+0.5)**2 + a2*(y-1)**2 - a2*b2
        while y >= 0:
            if d < 0:
                d += b2*(2*x+2) + a2*(-2*y+3)
                x += 1
            else:
                d += a2*(-2*y+3)
            y -= 1
            yield Point(x, y)
    for p in genFirst(round(radius[0]), round(radius[1])):
        yield center + p
        yield center - p
        yield center + Point(p.x, -p.y)
        yield center + Point(-p.x, p.y)


class EllipseGeometry(Geometry):
    def __init__(self, center: Optional[Point] = None, radius: Tuple[float, float] = (1, 1)) -> None:
        super().__init__()
        self.center = center if center else Point()
        self.radius = radius

    @staticmethod
    def fromRect(rect: Rect) -> EllipseGeometry:
        po, pv = rect.origin, rect.vertex()
        result = EllipseGeometry()
        result.center = (po+pv)*0.5
        result.radius = (rect.size.width/2, rect.size.height/2)
        return result

    @property
    def center(self) -> Point:
        return self._center

    @center.setter
    def center(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._center = value
        self.refreshBounds()

    @property
    def radius(self) -> Tuple[float, float]:
        return self._radius

    @radius.setter
    def radius(self, value: Tuple[float, float]) -> None:
        assert len(value) == 2
        self._radius = float(value[0]), float(value[1])
        self.refreshBounds()

    def transformed(self) -> Geometry:
        if self.transform is None:
            return self
        else:
            center, radius = self.center, self.radius
            lt = Point(center.x - radius[0], center.y - radius[1])
            rt = Point(center.x + radius[0], center.y - radius[1])
            lb = Point(center.x - radius[0], center.y + radius[1])
            rb = Point(center.x + radius[0], center.y + radius[1])
            lt, rt, lb, rb = self.transform.transform(lt), self.transform.transform(
                rt), self.transform.transform(lb), self.transform.transform(rb)
            minX, maxX = min(lt.x, rt.x, lb.x, rb.x), max(
                lt.x, rt.x, lb.x, rb.x)
            minY, maxY = min(lt.y, rt.y, lb.y, rb.y), max(
                lt.y, rt.y, lb.y, rb.y)
            return EllipseGeometry.fromRect(Rect.fromPoints(
                Point(minX, minY), Point(maxX, maxY)))

    def strokePoints(self, pen: Pen) -> Iterable[Point]:
        target = self
        if self.transform is not None:
            target = cast(EllipseGeometry, self.transformed())

        return _gen(target.center, target.radius)

    def fillPoints(self) -> Iterable[Point]:
        return []

    def _calculateBounds(self) -> Rect:
        target = self
        if self.transform is not None:
            target = cast(EllipseGeometry, self.transformed())
        lt = Point(target.center.x -
                   target.radius[0], target.center.y - target.radius[1])
        rb = Point(target.center.x +
                   target.radius[0], target.center.y + target.radius[1])
        return Rect.fromPoints(lt, rb)



from typing import Any

from PyQt5.QtCore import QModelIndex, Qt
from PyQt5.QtGui import QStandardItem, QStandardItemModel

from ImagingS import Color
from ImagingS.brush import Brush, SolidBrush
from ImagingS.Gui import icons


class BrushModel(QStandardItemModel):
    VALUE = 0

    def __init__(self, parent) -> None:
        super().__init__(0, 1, parent)
        self.setHeaderData(BrushModel.VALUE, Qt.Horizontal, "Value")

    def append(self, brush: Brush) -> None:
        fitem = None
        if isinstance(brush, SolidBrush):
            fitem = QStandardItem()
            self._setUserData(fitem, brush)
        if fitem:
            self.appendRow(fitem)
        else:
            raise Exception("Unsupport brush")

    def data(self, index: QModelIndex, role: int) -> Any:
        brush = self.getUserData(index)
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if isinstance(brush, SolidBrush):
                return brush.color.toHex()
        elif role == Qt.DecorationRole:
            if isinstance(brush, SolidBrush):
                return icons.getBrushIcon(brush)
        return super().data(index, role)

    def setData(self, index: QModelIndex, value: Any, role: int) -> bool:
        if role == Qt.EditRole:
            brush = self.getUserData(index)
            try:
                if isinstance(brush, SolidBrush):
                    brush.color = Color.fromHex(str(value))
            except Exception:
                return False
            self.dataChanged.emit(index, index, [role])
            return True
        return super().setData(index, value, role)

    def getUserData(self, index: QModelIndex) -> Brush:
        item = self.itemFromIndex(index)
        return item.data(Qt.UserRole)

    def _setUserData(self, item: QStandardItem, value: Brush) -> None:
        item.setData(value, Qt.UserRole)

    def clear_items(self) -> None:
        self.removeRows(0, self.rowCount())



from ImagingS import Colors, Point, Rect
from ImagingS.brush import SolidBrush


def test_solid_brush() -> None:
    c = SolidBrush(Colors.Black)
    assert c.colorAt(Point(), Rect()) == Colors.Black
    assert str(c).startswith("SolidBrush")



from __future__ import annotations

from typing import Optional

from ImagingS.brush import Brush, Brushes
from ImagingS.serialization import PropertySerializable


class Pen(PropertySerializable):
    def __init__(self, brush: Optional[Brush] = None, thickness: float = 1.0):
        super().__init__()
        self.thickness = 1.0
        self.brush = brush if brush else Brushes.Black

    @property
    def thickness(self) -> float:
        return self._thickness

    @thickness.setter
    def thickness(self, value: float) -> None:
        value = float(value)
        self._thickness = value

    @property
    def brush(self) -> Brush:
        return self._brush

    @brush.setter
    def brush(self, value: Brush) -> None:
        assert isinstance(value, Brush)
        self._brush = value



from __future__ import annotations

from typing import Dict, Optional, Tuple

import numpy as np

from ImagingS import Point
from ImagingS.transform import MatrixTransform


class ScaleTransform(MatrixTransform):
    def __init__(self, center: Optional[Point] = None, factor: Tuple[float, float] = (1, 1)) -> None:
        super().__init__()
        self.center = center if center else Point()
        self.factor = factor

    @property
    def center(self) -> Point:
        return self._center

    @center.setter
    def center(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._center = value

    @property
    def factor(self) -> Tuple[float, float]:
        return self._factor

    @factor.setter
    def factor(self, value: Tuple[float, float]) -> None:
        assert len(value) == 2
        self._factor = float(value[0]), float(value[1])
        self.matrix = np.array(
            [[self._factor[0], 0, 0],
             [0, self._factor[1], 0],
             [0, 0, 1]])

    def transform(self, origin: Point) -> Point:
        return super().transform(origin - self._center) + self.center

    def serialize(self) -> Dict:
        result = super().serialize()
        if MatrixTransform.S_Matrix in result:
            del result[MatrixTransform.S_Matrix]
        return result



from typing import Generic, Optional, TypeVar, Union

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QBrush, QColor, QKeyEvent, QPen
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsPathItem

from ImagingS.drawing import DrawingGroup
from ImagingS.geometry import Geometry
from ImagingS.transform import TransformGroup

from .. import Interactivity

_T = TypeVar("_T")


class TransformInteractivity(Interactivity, Generic[_T]):
    def __init__(self, target: Union[DrawingGroup, Geometry], transform: _T) -> None:
        super().__init__()
        self._target = target
        self._transform = transform

        self._viewItem = QGraphicsPathItem()
        pen = QPen(QColor("purple"))
        pen.setWidth(1)
        pen.setStyle(Qt.DashLine)
        pen.setCapStyle(Qt.RoundCap)
        self._viewItem.setPen(pen)
        self._viewItem.setBrush(QBrush(QColor("magenta")))

    @property
    def viewItem(self) -> Optional[QGraphicsItem]:
        return self._viewItem

    @property
    def target(self) -> Union[DrawingGroup, Geometry]:
        return self._target

    @property
    def transform(self) -> _T:
        return self._transform

    def start(self) -> None:
        self._oldTransform = self.target.transform
        if isinstance(self._oldTransform, TransformGroup):
            self._oldTransform.children.append(self.transform)
        else:
            self.target.transform = self.transform

        super().start()

    def end(self, success: bool) -> None:
        if not success:
            if isinstance(self._oldTransform, TransformGroup):
                del self._oldTransform.children[self.transform]
            self.target.transform = self._oldTransform
        super().end(success)

    def update(self) -> None:
        self.target.refreshBounds()
        super().update()

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Escape:
            self.end(False)
        super().onKeyPress(key)



from ._Pen import Pen
from ._RenderContext import (RenderContext,
                             NumpyArrayRenderContext, ProxyRenderContext, ClipRenderContext)
from ._Drawing import Drawing, DrawingGroup
from ._GeometryDrawing import GeometryDrawing

__all__ = (
    "Pen",
    "RenderContext",
    "Drawing",
    "DrawingGroup",
    "GeometryDrawing",
    "NumpyArrayRenderContext",
    "ProxyRenderContext",
    "ClipRenderContext"
)



from __future__ import annotations

from enum import Enum, unique
from typing import Iterable, Iterator, List, Optional, cast

from ImagingS import Point, Rect, RectMeasurer
from ImagingS.drawing import Pen

from . import Geometry


@unique
class CurveAlgorithm(Enum):
    Bezier = 0
    BSpline = 1


def _calcPointCount(points: List[Point]) -> int:
    cnt = 0
    for i in range(len(points)-1):
        cnt += abs(points[i+1]-points[i])
    cnt = int(cnt)
    return cnt


def _genBezier(points: List[Point]) -> Iterator[Point]:
    if len(points) < 2:
        return

    def casteljau(ps: List[Point], t: float) -> Point:
        cps = [p.clone() for p in ps]
        n = len(cps) - 1
        for i in range(1, n+1):
            for j in range(n-i+1):
                cps[j] = cps[j] + t * (cps[j+1] - cps[j])
        return cps[0]
    cnt = _calcPointCount(points)
    for t in range(cnt+1):
        p = casteljau(points, t / cnt)
        yield p


def _genBSpline3(points: List[Point]) -> Iterator[Point]:
    if len(points) < 4:
        return

    def subline(ps: List[Point]) -> Iterator[Point]:
        assert len(ps) == 4
        cnt = _calcPointCount(ps)
        for tc in range(cnt+1):
            t = tc / cnt
            t2 = t * t
            t3 = t2 * t
            f03 = (3*t2 - t3 - 3*t + 1) / 6
            f13 = (3*t3 - 6*t2 + 4) / 6
            f23 = (3*t2 - 3*t3 + 3*t + 1) / 6
            f33 = t3 / 6
            yield f03*ps[0] + f13*ps[1] + f23*ps[2] + f33*ps[3]
    for i in range(len(points)-3):
        for p in subline(points[i:i+4]):
            yield p


class CurveGeometry(Geometry):
    def __init__(self, controlPoints: Optional[List[Point]] = None, algorithm: Optional[CurveAlgorithm] = None) -> None:
        super().__init__()
        self.controlPoints = controlPoints if controlPoints else []
        self.algorithm = algorithm if algorithm else CurveAlgorithm.Bezier

    @property
    def controlPoints(self) -> List[Point]:
        return self._controlPoints

    @controlPoints.setter
    def controlPoints(self, value: List[Point]) -> None:
        assert isinstance(value, list)
        self._controlPoints = value
        self.refreshBounds()

    @property
    def algorithm(self) -> CurveAlgorithm:
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: CurveAlgorithm) -> None:
        assert isinstance(value, CurveAlgorithm)
        self._algorithm = value

    def transformed(self) -> Geometry:
        if self.transform is None:
            return self
        else:
            vertexes = list(map(self.transform.transform, self.controlPoints))
            return CurveGeometry(vertexes, self.algorithm)

    def strokePoints(self, pen: Pen) -> Iterable[Point]:
        target = self
        if self.transform is not None:
            target = cast(CurveGeometry, self.transformed())
        gen = None
        if target.algorithm is CurveAlgorithm.Bezier:
            gen = _genBezier
        elif target.algorithm is CurveAlgorithm.BSpline:
            gen = _genBSpline3
        return gen(target.controlPoints)

    def fillPoints(self) -> Iterable[Point]:
        return []

    def _calculateBounds(self) -> Rect:
        target = self
        if self.transform is not None:
            target = cast(CurveGeometry, self.transformed())
        measure = RectMeasurer()
        measure.append(target.controlPoints)
        return measure.result()



from __future__ import annotations

import uuid
from typing import Collection, Dict, Iterator, List, Optional, TypeVar, Union

from ImagingS.serialization import PropertySerializable


class IdObject:
    def __init__(self, id: str = "") -> None:
        super().__init__()
        self.id = id if id else str(uuid.uuid1())

    @property
    def id(self) -> str:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        self._id = str(value)

    def parent(self) -> Optional[IdObjectList]:
        if hasattr(self, "_parent"):
            return self._parent
        return None

    def setParent(self, value: Optional[IdObjectList]) -> None:
        if value is None:
            if hasattr(self, "_parent"):
                del self._parent
        else:
            self._parent = value


_T = TypeVar("_T", bound=IdObject)


class IdObjectList(PropertySerializable, Collection[_T]):
    def __init__(self) -> None:
        super().__init__()
        self.items = []

    @property
    def items(self) -> List[_T]:
        return self._items

    @items.setter
    def items(self, value: List[_T]) -> None:
        if hasattr(self, "_items") and self._items is value:
            return
        self._items: List[_T] = []
        self._ids: Dict[str, _T] = {}
        for item in value:
            self.append(item)

    def append(self, item: _T) -> None:
        if item.id in self:
            raise Exception(f"The id '{item.id}' has been added.")
        item.setParent(self)
        self._items.append(item)
        self._ids[item.id] = item

    def clear(self) -> None:
        self.items = []

    def setItemId(self, oldId, newId) -> None:
        if oldId not in self._ids:
            return
        if newId in self._ids:
            raise Exception(f"The id '{newId}' has been added.")
        item = self._ids[oldId]
        del self._ids[oldId]
        item.id = newId
        self._ids[item.id] = item

    def __contains__(self, item: Union[str, IdObject]) -> bool:
        if isinstance(item, str):
            return item in self._ids
        else:
            return item.id in self._ids

    def __delitem__(self, key: Union[str, int, IdObject]) -> None:
        if isinstance(key, IdObject):
            key = key.id
        elif isinstance(key, int):
            key = self._items[key].id
        if key not in self:
            return
        item = self._ids[key]
        item.setParent(None)
        self._items.remove(item)
        del self._ids[key]

    def __getitem__(self, key: Union[str, int]) -> _T:
        if isinstance(key, int):
            key = self._items[key].id
        return self._ids[key]

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[_T]:
        return iter(self._items)



from ._Brush import Brush
from ._Solid import SolidBrush
from ImagingS import Colors

__all__ = (
    "Brush",
    "SolidBrush",
    "Brushes"
)


class Brushes:
    Black = SolidBrush(Colors.Black)
    White = SolidBrush(Colors.White)
    Red = SolidBrush(Colors.Red)
    Blue = SolidBrush(Colors.Blue)
    Green = SolidBrush(Colors.Green)



from enum import Enum, unique
from typing import Optional

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

import ImagingS.Gui.ui as ui
from ImagingS.document import Document

from . import CodePage, CodePageState, VisualPage, VisualPageState


@unique
class DocumentEditorState(Enum):
    Disable = 0
    Visual = 1
    Code = 2


class DocumentEditor(QWidget, ui.DocumentEditor):
    documentCommitted = pyqtSignal(Document, str)
    stateChanged = pyqtSignal()
    messaged = pyqtSignal(str)

    def __init__(self) -> None:
        super().__init__()
        self.setupUi(self)
        self.setupVisual()
        self.setupCode()

        self.code.documentCommitted.connect(self.code_documentCommitted)
        self.visual.documentCommitted.connect(self.visual_documentCommitted)

        self.setEnabled(False)

    def setupCode(self) -> None:
        self.code = CodePage()
        self.code.setObjectName("code")
        self.grdCode.addWidget(self.code, 0, 0, 1, 1)
        self.code.messaged.connect(self.visualCode_messaged)

    def setupVisual(self) -> None:
        self.visual = VisualPage()
        self.visual.setObjectName("visual")
        self.grdVisual.addWidget(self.visual, 0, 0, 1, 1)
        self.visual.messaged.connect(self.visualCode_messaged)

    @property
    def document(self) -> Optional[Document]:
        if hasattr(self, "_document"):
            return self._document
        else:
            return None

    @property
    def state(self) -> DocumentEditorState:
        if hasattr(self, "_document"):
            if self.swgMain.currentIndex() == 0:
                return DocumentEditorState.Visual
            else:
                return DocumentEditorState.Code
        else:
            return DocumentEditorState.Disable

    def enable(self, doc: Document) -> None:
        assert self.state is DocumentEditorState.Disable
        self.setEnabled(True)
        self._document = doc
        self.switchVisual()

    def disable(self) -> None:
        assert self.state is not DocumentEditorState.Disable
        if hasattr(self, "_document"):
            del self._document
        if self.visual.state is not VisualPageState.Disable:
            self.visual.disable()
        if self.code.state is not CodePageState.Disable:
            self.code.disable()
        self.setEnabled(False)
        self.stateChanged.emit()

    def fresh(self) -> None:
        state = self.state
        if state is DocumentEditorState.Disable:
            return
        elif state is DocumentEditorState.Visual:
            self.visual.fresh()
        elif state is DocumentEditorState.Code:
            self.code.fresh()

    def switchCode(self) -> bool:
        if self.document is None:
            return True
        if self.code.state is not CodePageState.Disable:
            return True
        if self.visual.state is VisualPageState.Normal:
            self.visual.disable()
        elif self.visual.state is VisualPageState.Interactive:
            self.messaged.emit("Still in interactive mode.")
            return False
        self.code.enable(self.document)
        self.swgMain.setCurrentIndex(1)
        self.stateChanged.emit()
        return True

    def switchVisual(self) -> bool:
        if self.document is None:
            return True
        if self.visual.state is not VisualPageState.Disable:
            return True
        if self.code.state is CodePageState.Normal:
            self.code.disable()
        elif self.code.state is CodePageState.Changed:
            self.messaged.emit("Code changes not committed.")
            return False
        self.visual.enable(self.document)
        self.swgMain.setCurrentIndex(0)
        self.stateChanged.emit()
        return True

    def visualCode_messaged(self, message: str) -> None:
        self.messaged.emit(message)

    def code_documentCommitted(self, doc: Document, message: str):
        self._document = doc
        self.documentCommitted.emit(doc, message)

    def visual_documentCommitted(self, doc: Document, message: str):
        self._document = doc
        self.documentCommitted.emit(doc, message)



from ._Property import Property, getProperties
from ._Serializable import PropertySerializable, Serializable

__all__ = (
    "Property",
    "getProperties",
    "Serializable",
    "PropertySerializable"
)



from __future__ import annotations

from typing import Dict, Optional, cast

from ImagingS import Point, Rect

from . import LineAlgorithm, PolygonGeometry


class RectangleGeometry(PolygonGeometry):
    def __init__(self, rect: Optional[Rect] = None, algorithm: Optional[LineAlgorithm] = None) -> None:
        super().__init__(algorithm=algorithm)
        self.rect = rect if rect else Rect()

    @property
    def rect(self) -> Rect:
        return self._rect

    @rect.setter
    def rect(self, value: Rect) -> None:
        assert isinstance(value, Rect)
        self._rect = value
        self.vertexes = [
            self._rect.origin,
            self._rect.origin + Point(self._rect.size.width, 0),
            self._rect.origin +
            Point(self._rect.size.width, self._rect.size.height),
            self._rect.origin + Point(0, self._rect.size.height),
        ]

    def serialize(self) -> Dict:
        result = super().serialize()
        if PolygonGeometry.S_Vertexes in result:
            del result[PolygonGeometry.S_Vertexes]
        return result

    def _calculateBounds(self) -> Rect:
        if self.transform is not None:
            target = cast(PolygonGeometry, self.transformed())
            return target._calculateBounds()
        else:
            return self.rect



from __future__ import annotations

from typing import Dict, Optional

import numpy as np

from ImagingS import Point

from . import MatrixTransform


class TranslateTransform(MatrixTransform):
    def __init__(self, delta: Optional[Point] = None) -> None:
        super().__init__()
        self.delta = delta if delta else Point()

    @property
    def delta(self) -> Point:
        return self._delta

    @delta.setter
    def delta(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._delta = value
        self.matrix = np.array(
            [[1, 0, 0],
             [0, 1, 0],
             [self._delta.x, self._delta.y, 1]])

    def transform(self, origin: Point) -> Point:
        return origin + self._delta

    def serialize(self) -> Dict:
        result = super().serialize()
        if MatrixTransform.S_Matrix in result:
            del result[MatrixTransform.S_Matrix]
        return result



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/runner/work/ImagingS/ImagingS/src/ImagingS/Gui/uic/DocumentEditor.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DocumentEditor(object):
    def setupUi(self, DocumentEditor):
        DocumentEditor.setObjectName("DocumentEditor")
        DocumentEditor.resize(400, 300)
        self.gridLayout = QtWidgets.QGridLayout(DocumentEditor)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.swgMain = QtWidgets.QStackedWidget(DocumentEditor)
        self.swgMain.setObjectName("swgMain")
        self.pagVisual = QtWidgets.QWidget()
        self.pagVisual.setGeometry(QtCore.QRect(0, 0, 400, 254))
        self.pagVisual.setObjectName("pagVisual")
        self.grdVisual = QtWidgets.QGridLayout(self.pagVisual)
        self.grdVisual.setContentsMargins(0, 0, 0, 0)
        self.grdVisual.setObjectName("grdVisual")
        self.swgMain.addWidget(self.pagVisual)
        self.pagCode = QtWidgets.QWidget()
        self.pagCode.setGeometry(QtCore.QRect(0, 0, 400, 254))
        self.pagCode.setObjectName("pagCode")
        self.grdCode = QtWidgets.QGridLayout(self.pagCode)
        self.grdCode.setContentsMargins(0, 0, 0, 0)
        self.grdCode.setObjectName("grdCode")
        self.swgMain.addWidget(self.pagCode)
        self.gridLayout.addWidget(self.swgMain, 0, 0, 1, 1)

        self.retranslateUi(DocumentEditor)
        self.swgMain.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(DocumentEditor)

    def retranslateUi(self, DocumentEditor):
        _translate = QtCore.QCoreApplication.translate
        DocumentEditor.setWindowTitle(_translate("DocumentEditor", "Editor"))



from __future__ import annotations

from typing import Iterable, Optional, Union

from ImagingS.serialization import PropertySerializable
from ImagingS import Point, Size


class Rect(PropertySerializable):
    _infinite: Optional[Rect] = None

    def __init__(self, origin: Optional[Point] = None, size: Optional[Size] = None) -> None:
        super().__init__()
        self.origin = origin if origin else Point()
        self.size = size if size else Size()

    @classmethod
    def infinite(cls) -> Rect:
        if cls._infinite is None:
            cls._infinite = Rect.fromPoints(Point(
                float("-inf"), float("-inf")), Point(float("inf"), float("inf")))
        return cls._infinite

    @staticmethod
    def fromPoints(p1: Point, p2: Point) -> Rect:
        x1, y1 = p1.asTuple()
        x2, y2 = p2.asTuple()
        xmin, xmax = min(x1, x2), max(x1, x2)
        ymin, ymax = min(y1, y2), max(y1, y2)
        return Rect(Point(xmin, ymin), Size(xmax - xmin, ymax - ymin))

    def __eq__(self, obj: Rect) -> bool:
        return self.origin == obj.origin and self.size == obj.size

    def __repr__(self) -> str:
        return f"Rect({self.origin}, {self.size})"

    def __contains__(self, point: Point) -> bool:
        delta = point - self.origin
        return 0 <= delta.x <= self.size.width and 0 <= delta.y <= self.size.height

    @property
    def origin(self) -> Point:
        return self._origin

    @origin.setter
    def origin(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._origin = value

    @property
    def size(self) -> Size:
        return self._size

    @size.setter
    def size(self, value: Size) -> None:
        assert isinstance(value, Size)
        self._size = value

    def vertex(self) -> Point:
        return Point(self.origin.x+self.size.width, self.origin.y+self.size.height)


class RectMeasurer:
    def __init__(self) -> None:
        super().__init__()
        self._lx = float("inf")
        self._ly = float("inf")
        self._rx = float("-inf")
        self._ry = float("-inf")

    def result(self) -> Rect:
        result = Rect.fromPoints(Point(
            self._lx, self._ly), Point(self._rx, self._ry))
        return result

    def append(self, position: Union[Point, Iterable[Point]]) -> None:
        if isinstance(position, Point):
            x, y = position.asTuple()
            self._lx = min(self._lx, x)
            self._ly = min(self._ly, y)
            self._rx = max(self._rx, x)
            self._ry = max(self._ry, y)
        else:
            for p in position:
                self.append(p)
